package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MaterialTypeRepository
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class MaterialTypeServiceTest :
  AbstractExternalNamedModelServiceTest<MaterialType, MaterialTypeSaveDto, MaterialTypeUpdateDto, MaterialTypeService, MaterialTypeRepository>() {
    override val repository: MaterialTypeRepository = mock()
    private val materialService: MaterialService = mock()
    override val service: MaterialTypeService = spy(MaterialTypeServiceImpl(repository, materialService))
    override val entity: MaterialType = materialType(id = 0L, name = "material type", prefix = "MAT")
    override val anotherEntity: MaterialType = materialType(id = 1L, name = "another material type", prefix = "AMT")
    override val entityWithEntityName: MaterialType = materialType(2L, name = entity.name, prefix = "EEN")
    private val systemType = materialType(id = 3L, name = "systype", prefix = "SYS", systemType = true)
    private val anotherSystemType = materialType(id = 4L, name = "another systype", prefix = "ASY", systemType = true)
    override val entitySaveDto: MaterialTypeSaveDto = spy(materialTypeSaveDto(name = "material type", prefix = "MAT"))
    override val entityUpdateDto: MaterialTypeUpdateDto =
      spy(materialTypeUpdateDto(id = 0L, name = "material type", prefix = "MAT"))

    @AfterEach
    override fun afterEach() {
        reset(materialService)
        super.afterEach()
    }

    // existsByPrefix()

    @Test
    fun `existsByPrefix() returns true when a material type with the given prefix exists`() {
        whenever(repository.existsByPrefix(entity.prefix)).doReturn(true)

        val found = service.existsByPrefix(entity.prefix)

        assertTrue(found)
    }

    @Test
    fun `existsByPrefix() returns false when no material type with the given prefix exists`() {
        whenever(repository.existsByPrefix(entity.prefix)).doReturn(false)

        val found = service.existsByPrefix(entity.prefix)

        assertFalse(found)
    }

    // isUsedByMaterial()

    @Test
    fun `isUsedByMaterial() returns true when materialService_existsByMaterialType() returns true`() {
        whenever(materialService.existsByMaterialType(entity)).doReturn(true)

        val found = service.isUsedByMaterial(entity)

        assertTrue(found)
    }

    @Test
    fun `isUsedByMaterial() returns false when materialService_existsByMaterialType() returns false`() {
        whenever(materialService.existsByMaterialType(entity)).doReturn(false)

        val found = service.isUsedByMaterial(entity)

        assertFalse(found)
    }

    // getAllSystemTypes()

    @Test
    fun `getAllSystemTypes() returns all system types`() {
        whenever(repository.findAllBySystemTypeIs(true)).doReturn(listOf(systemType, anotherSystemType))

        val found = service.getAllSystemTypes()

        assertTrue(found.contains(systemType))
        assertTrue(found.contains(anotherSystemType))
    }

    // getAllNonSystemTypes()

    @Test
    fun `getAllNonSystemTypes() returns all non system types`() {
        whenever(repository.findAllBySystemTypeIs(false)).doReturn(listOf(entity, anotherEntity))

        val found = service.getAllNonSystemType()

        assertTrue(found.contains(entity))
        assertTrue(found.contains(anotherEntity))
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, service)
    }

    // saveMaterialType()

    @Test
    fun `saveMaterialType() throws AlreadyExistsException when a material type with the given prefix already exists`() {
        doReturn(true).whenever(service).existsByPrefix(entity.prefix)

        assertThrows<AlreadyExistsException> { service.save(entity) }
          .assertErrorCode("prefix")
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
      withBaseUpdateDtoTest(entity, entityUpdateDto, service, { any() })

    override fun `update() saves in the repository and returns the updated value`() {
        whenever(repository.save(entity)).doReturn(entity)
        whenever(repository.findByName(entity.name)).doReturn(null)
        whenever(repository.findByPrefix(entity.prefix)).doReturn(null)
        doReturn(true).whenever(service).existsById(entity.id!!)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.update(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    override fun `update() throws NotFoundException when no entity with the given id exists in the repository`() {
        whenever(repository.findByName(entity.name)).doReturn(null)
        whenever(repository.findByPrefix(entity.prefix)).doReturn(null)
        doReturn(false).whenever(service).existsById(entity.id!!)
        doReturn(null).whenever(service).getById(entity.id!!)

        assertThrows<NotFoundException> { service.update(entity) }
          .assertErrorCode()
    }

    override fun `update() throws AlreadyExistsException when an entity with the updated name exists`() {
        whenever(repository.findByName(entity.name)).doReturn(entityWithEntityName)
        whenever(repository.findByPrefix(entity.prefix)).doReturn(null)
        doReturn(true).whenever(service).existsById(entity.id!!)
        doReturn(entity).whenever(service).getById(entity.id!!)

        assertThrows<AlreadyExistsException> { service.update(entity) }
          .assertErrorCode("name")
    }

    @Test
    fun `update() throws AlreadyExistsException when an entity with the updated prefix exists`() {
        val anotherMaterialType = materialType(prefix = entity.prefix)
        whenever(repository.findByPrefix(entity.prefix)).doReturn(anotherMaterialType)
        doReturn(entity).whenever(service).getById(entity.id!!)

        assertThrows<AlreadyExistsException> { service.update(entity) }
          .assertErrorCode("prefix")
    }

    // delete()

    override fun `delete() deletes in the repository`() {
        whenCanBeDeleted {
            super.`delete() deletes in the repository`()
        }
    }

    override fun `deleteById() deletes the entity with the given id in the repository`() {
        whenCanBeDeleted {
            super.`deleteById() deletes the entity with the given id in the repository`()
        }
    }

    private fun whenCanBeDeleted(id: Long = any(), test: () -> Unit) {
        whenever(repository.canBeDeleted(id)).doReturn(true)

        test()
    }
}
